//============================================================================
// 635_WinTest Code.
// Copyright 2001-2004, Crystalfontz America, Inc. Written by Brent A. Crosby
//
// You may use this code in any project (including commercial) as long as
// at least one of the intents of the project is to communicate with a
// Crystalfontz display.
//
// http://www.crystalfontz.com brent@crystalfontz.com
//
// !!!!!!!! THIS IS TEST CODE ONLY !!!!!!!!
// !!!!!!!!  USE AT YOUR OWN RISK  !!!!!!!!
//
//============================================================================
#include "stdafx.h"
#include "serial.h"
//============================================================================
//Only included to access serial port status field in main window
#include "635_wintest.h"
#include "GPIO.h"
#include "KeyPads.h"
#include "TempsDlg.h"
#include "Fans.h"
#include "Packet_Debug.h"
#include "Live_Display.h"
#include "LED_Editor.h"
#include "635_WinTestDlg.h"
//------------------------------------------------------------------------------
HANDLE
  hCom=NULL;
CWnd
  *OurParent=NULL;
/*---------------------------------------------------------------------------*/
//This is some code that kind of makes the windows stuff look a little
//like the DOS/635 interrupt driven serial stuff. Basically there is a
//circular buffer, and Sync_Read_Buffer() uses ReadFile() to put data
//into the circular buffer much like the DOS stuff uses an ISR to put
//the data into the buffer. Then the rest of the functions work like
//the counterparts in the 635,
#define RECEIVEBUFFERSIZE    8192
ubyte
  SerialReceiveBuffer[RECEIVEBUFFERSIZE];
DWORD
  ReceiveBufferHead;
DWORD
  ReceiveBufferTail;
DWORD
  ReceiveBufferTailPeek;
/*---------------------------------------------------------------------------*/
void EmptyReceiveBuffer(void)
  {
  ReceiveBufferHead=ReceiveBufferTail=0;
  }
//------------------------------------------------------------------------------
int Serial_Init(char *port, int baud_rate, CWnd* pParent)
  {
  DCB
    dcb;
  BOOL
    fSuccess;

  OurParent=pParent;
  if(hCom)
    CloseHandle(hCom);

  hCom = CreateFile(
    port,
    GENERIC_READ | GENERIC_WRITE,
    0,             /* comm devices must be opened w/exclusive-access */
    NULL,          /* no security attrs */
    OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
    0,             /* not overlapped I/O */
    NULL           /* hTemplate must be NULL for comm devices */
    );

  if(hCom == INVALID_HANDLE_VALUE)
    {
    ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("CreateFile() failed!");
    hCom=0;
    return(1);
    }

  /*
   * Omit the call to SetupComm to use the default queue sizes.
   * Get the current configuration.
   */
  fSuccess = GetCommState(hCom, &dcb);
  if(!fSuccess)
    {
    ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("GetCommState() failed!");
    hCom=0;
    return(2);
    }

  /* Fill in the DCB: baud=x, 8 data bits, no parity, 1 stop bit. */
  dcb.BaudRate = baud_rate;
  dcb.ByteSize = 8;
  dcb.Parity = NOPARITY;
  dcb.StopBits = ONESTOPBIT;
  dcb.fRtsControl = RTS_CONTROL_ENABLE;
  dcb.fDtrControl = DTR_CONTROL_ENABLE; 
  dcb.fAbortOnError = FALSE; 
  fSuccess = SetCommState(hCom, &dcb);

  if(!fSuccess)
    {
    hCom=0;
    ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("SetCommState() failed!");
    return(3);
    }

  COMMTIMEOUTS
    our_timeouts;

  if(TRUE!=GetCommTimeouts(
      hCom,             // handle of communications device
      &our_timeouts))   // address of comm. time-outs structure
    {
    hCom=0;
    ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("GetCommTimeouts() failed!");
    return(4);
    }

  //9600 is 1 byte per mS. Set it to 2 mS so the read will stay
  //as long as data is coming in. Will wait at least 100mS for
  //the 635 to respond.
  our_timeouts.ReadIntervalTimeout=2000;
  our_timeouts.ReadTotalTimeoutMultiplier=2000;  // 1 second
  our_timeouts.ReadTotalTimeoutConstant=2000;

  if(TRUE!=SetCommTimeouts(
      hCom,           // handle of communications device
      &our_timeouts))   // address of comm. time-outs structure
    {
    hCom=0;
    ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("SetCommTimeouts() failed!");
    return(5);
    }

  EmptyReceiveBuffer();

  return(0);
  }
//------------------------------------------------------------------------------
void Uninit_Serial()
  {
  CloseHandle(hCom);
  hCom=NULL;
  }
//------------------------------------------------------------------------------
//Gets incoming data from Window's ReadFile() and puts it into
//SerialReceiveBuffer[].
void Sync_Read_Buffer(void)
  {
  ubyte
    Incoming[8192];
  DWORD
    BytesRead;
  COMSTAT
    status;
  DWORD
    errors;

  if(!hCom)
    return;

  //Find out how much data is available to be read.
  ClearCommError(
    hCom,      // handle to communications device
    &errors,   // pointer to variable to receive error codes
    &status);  // pointer to buffer for communications status
 
  if(status.cbInQue)
    {
    if(!ReadFile(hCom,               // handle of file to read
                 (LPVOID *)Incoming, // pointer to buffer that receives data
                 status.cbInQue,     // number of bytes to read
                 &BytesRead,         // pointer to number of bytes read
                 NULL))              // Overlaped structure address
      ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("ReadFile() failed!");
    else
      //Read the incoming ubyte, store it,
      for(DWORD i=0;i<BytesRead;i++)
        {
        SerialReceiveBuffer[ReceiveBufferHead]=Incoming[i];
        //Increment the pointer and wrap if needed.
        ReceiveBufferHead++;
        if(RECEIVEBUFFERSIZE<=ReceiveBufferHead)
          ReceiveBufferHead=0;
        }
    }
  }
/*---------------------------------------------------------------------------*/
DWORD BytesAvail(void)
  {
  //LocalReceiveBufferHead and return_value are a signed variables.
  register int
    LocalReceiveBufferHead;
  register int
    return_value;
    
  //Get a register copy of the head pointer, since an interrupt that
  //modifies the head pointer could screw up the value. Convert it to
  //our signed local variable.
  LocalReceiveBufferHead=ReceiveBufferHead;
  if((return_value=(LocalReceiveBufferHead-(int)ReceiveBufferTail))<0)
    return_value+=RECEIVEBUFFERSIZE;
  return(return_value);
  }
/*---------------------------------------------------------------------------*/
ubyte GetByte(void)
  {
  register DWORD
    LocalReceiveBufferTail;
  register DWORD
    LocalReceiveBufferHead;
  register ubyte
    return_byte;

  //Get a register copy of the tail pointer for speed and size.
  LocalReceiveBufferTail=ReceiveBufferTail;
  
  //Get a register copy of the head pointer, since an interrupt that
  //modifies the head pointer could screw up the value.
  LocalReceiveBufferHead=ReceiveBufferHead;
  
  //See if there are any more bytes available.
  if(LocalReceiveBufferTail!=LocalReceiveBufferHead)
    {
    //There is at least one more ubyte.
    return_byte=SerialReceiveBuffer[LocalReceiveBufferTail];
    //Increment the pointer and wrap if needed.
    LocalReceiveBufferTail++;
    if(RECEIVEBUFFERSIZE<=LocalReceiveBufferTail)
      LocalReceiveBufferTail=0;
    //Update the real ReceiveBufferHead with our register copy. Make sure
    //the ISR does not get serviced during the transfer.
    ReceiveBufferTail=LocalReceiveBufferTail;
    }
  return(return_byte);
  }
/*---------------------------------------------------------------------------*/
DWORD PeekBytesAvail(void)
  {
  //LocalReceiveBufferHead and return_value are a signed variables.
  register int
    LocalReceiveBufferHead;
  register int
    return_value;
    
  //Get a register copy of the head pointer, since an interrupt that
  //modifies the head pointer could screw up the value. Convert it to
  //our signed local variable.
  LocalReceiveBufferHead=ReceiveBufferHead;
  if((return_value=(LocalReceiveBufferHead-(int)ReceiveBufferTailPeek))<0)
    return_value+=RECEIVEBUFFERSIZE;
  return(return_value);
  }
/*---------------------------------------------------------------------------*/
void Sync_Peek_Pointer(void)
  {
  ReceiveBufferTailPeek=ReceiveBufferTail;
  }
/*---------------------------------------------------------------------------*/
void AcceptPeekedData(void)
  {
  ReceiveBufferTail=ReceiveBufferTailPeek;
  }
/*---------------------------------------------------------------------------*/
ubyte PeekByte(void)
  {
  register int
    LocalReceiveBufferTailPeek;
  register int
    LocalReceiveBufferHead;
  register ubyte
    return_byte;

  //Get a register copy of the tail pointer for speed and size.
  LocalReceiveBufferTailPeek=ReceiveBufferTailPeek;
  
  //Get a register copy of the head pointer, since an interrupt that
  //modifies the head pointer could screw up the value.
  LocalReceiveBufferHead=ReceiveBufferHead;
  
  //See if there are any more bytes available.
  if(LocalReceiveBufferTailPeek!=LocalReceiveBufferHead)
    {
    //There is at least one more ubyte.
    return_byte=SerialReceiveBuffer[LocalReceiveBufferTailPeek];
    //Increment the pointer and wrap if needed.
    LocalReceiveBufferTailPeek++;
    if(RECEIVEBUFFERSIZE<=LocalReceiveBufferTailPeek)
      LocalReceiveBufferTailPeek=0;
    //Update the real ReceiveBufferHead with our register copy. Make sure
    //the ISR does not get serviced during the transfer.
    ReceiveBufferTailPeek=LocalReceiveBufferTailPeek;
    }
  return(return_byte);
  }
/*---------------------------------------------------------------------------*/ 
#define CLEAN_CRC_SOURCE 0

#if CLEAN_CRC_SOURCE

word get_crc(ubyte *bufptr,word len)
  {
  register unsigned int
    newCRC;
  //Put the current byte in here.
  ubyte
    data;
  int
    bit_count;
  //This seed makes the output of this shift based algorithm match
  //the table based algorithm. The center 16 bytes of the 32-bit
  //"newCRC" are used for the CRC. The MSb of the lower byte is used
  //to see what bit was shifted out of the center 16 bit CRC
  //accumulator ("carry flag analog");
  newCRC=0x00F32100;
  while(len--)
    {
    //Get the next byte in the stream.
    data=*bufptr++;
    //Push this byte's bits through a software
    //implementation of a hardware shift & xor. 
    for(bit_count=0;bit_count<=7;bit_count++)
      {
      //Shift the CRC accumulator
      newCRC>>=1;

      //The new MSB of the CRC accumulator comes
      //from the LSB of the current data byte.
      if(data&0x01)
        newCRC|=0x00800000;

      //If the low bit of the current CRC accumulator was set
      //before the shift, then we need to XOR the accumulator
      //with the polynomial (center 16 bits of 0x00840800)
      if(newCRC&0x00000080)
        newCRC^=0x00840800;
      //Shift the data byte to put the next bit of the stream
      //into position 0.
      data>>=1;
      }
    }

  //All the data has been done. Do 16 more bits of 0 data.
  for(bit_count=0;bit_count<=15;bit_count++)
    {
    //Shift the CRC accumulator
    newCRC>>=1;

    //If the low bit of the current CRC accumulator was set
    //before the shift we need to XOR the accumulator with
    //0x00840800.
    if(newCRC&0x00000080)
      newCRC^=0x00840800;
    }
  //Return the center 16 bits, making this CRC match the one's
  //complement that is sent in the packet.
  return((~newCRC)>>8);
  }

#else
//This code is from the IRDA LAP documentation, which appears to 
//have been copied from PPP:
//
// http://irda.affiniscape.com/associations/2494/files/Specifications/IrLAP11_Plus_Errata.zip
//
//I doubt that there are any worries about the legality of this code,
//searching for the first line of the table below, it appears that
//the code is already included in the linux 2.6 kernel "Driver for
//ST5481 USB ISDN modem". This is an "industry standard" algorithim
//and I do not think there are ANY issues with it at all.
word get_crc(ubyte *bufptr,word len)
  {
  //CRC lookup table to avoid bit-shifting loops.
  static const word crcLookupTable[256] =
    {0x00000,0x01189,0x02312,0x0329B,0x04624,0x057AD,0x06536,0x074BF,
     0x08C48,0x09DC1,0x0AF5A,0x0BED3,0x0CA6C,0x0DBE5,0x0E97E,0x0F8F7,
     0x01081,0x00108,0x03393,0x0221A,0x056A5,0x0472C,0x075B7,0x0643E,
     0x09CC9,0x08D40,0x0BFDB,0x0AE52,0x0DAED,0x0CB64,0x0F9FF,0x0E876,
     0x02102,0x0308B,0x00210,0x01399,0x06726,0x076AF,0x04434,0x055BD,
     0x0AD4A,0x0BCC3,0x08E58,0x09FD1,0x0EB6E,0x0FAE7,0x0C87C,0x0D9F5,
     0x03183,0x0200A,0x01291,0x00318,0x077A7,0x0662E,0x054B5,0x0453C,
     0x0BDCB,0x0AC42,0x09ED9,0x08F50,0x0FBEF,0x0EA66,0x0D8FD,0x0C974,
     0x04204,0x0538D,0x06116,0x0709F,0x00420,0x015A9,0x02732,0x036BB,
     0x0CE4C,0x0DFC5,0x0ED5E,0x0FCD7,0x08868,0x099E1,0x0AB7A,0x0BAF3,
     0x05285,0x0430C,0x07197,0x0601E,0x014A1,0x00528,0x037B3,0x0263A,
     0x0DECD,0x0CF44,0x0FDDF,0x0EC56,0x098E9,0x08960,0x0BBFB,0x0AA72,
     0x06306,0x0728F,0x04014,0x0519D,0x02522,0x034AB,0x00630,0x017B9,
     0x0EF4E,0x0FEC7,0x0CC5C,0x0DDD5,0x0A96A,0x0B8E3,0x08A78,0x09BF1,
     0x07387,0x0620E,0x05095,0x0411C,0x035A3,0x0242A,0x016B1,0x00738,
     0x0FFCF,0x0EE46,0x0DCDD,0x0CD54,0x0B9EB,0x0A862,0x09AF9,0x08B70,
     0x08408,0x09581,0x0A71A,0x0B693,0x0C22C,0x0D3A5,0x0E13E,0x0F0B7,
     0x00840,0x019C9,0x02B52,0x03ADB,0x04E64,0x05FED,0x06D76,0x07CFF,
     0x09489,0x08500,0x0B79B,0x0A612,0x0D2AD,0x0C324,0x0F1BF,0x0E036,
     0x018C1,0x00948,0x03BD3,0x02A5A,0x05EE5,0x04F6C,0x07DF7,0x06C7E,
     0x0A50A,0x0B483,0x08618,0x09791,0x0E32E,0x0F2A7,0x0C03C,0x0D1B5,
     0x02942,0x038CB,0x00A50,0x01BD9,0x06F66,0x07EEF,0x04C74,0x05DFD,
     0x0B58B,0x0A402,0x09699,0x08710,0x0F3AF,0x0E226,0x0D0BD,0x0C134,
     0x039C3,0x0284A,0x01AD1,0x00B58,0x07FE7,0x06E6E,0x05CF5,0x04D7C,
     0x0C60C,0x0D785,0x0E51E,0x0F497,0x08028,0x091A1,0x0A33A,0x0B2B3,
     0x04A44,0x05BCD,0x06956,0x078DF,0x00C60,0x01DE9,0x02F72,0x03EFB,
     0x0D68D,0x0C704,0x0F59F,0x0E416,0x090A9,0x08120,0x0B3BB,0x0A232,
     0x05AC5,0x04B4C,0x079D7,0x0685E,0x01CE1,0x00D68,0x03FF3,0x02E7A,
     0x0E70E,0x0F687,0x0C41C,0x0D595,0x0A12A,0x0B0A3,0x08238,0x093B1,
     0x06B46,0x07ACF,0x04854,0x059DD,0x02D62,0x03CEB,0x00E70,0x01FF9,
     0x0F78F,0x0E606,0x0D49D,0x0C514,0x0B1AB,0x0A022,0x092B9,0x08330,
     0x07BC7,0x06A4E,0x058D5,0x0495C,0x03DE3,0x02C6A,0x01EF1,0x00F78};

  register word
    newCrc;
  newCrc=0xFFFF;
  //This algorithim is based on the IrDA LAP example.
  while(len--)
    newCrc = (newCrc >> 8) ^ crcLookupTable[(newCrc ^ *bufptr++) & 0xff];

  //Make this crc match the one's complement that is sent in the packet.
  return(~newCrc);
  }
#endif
//============================================================================

COMMAND_PACKET
  incoming_response;

//============================================================================
//                              send_packet()
//
// send_packet() will send set the CRC in outgoing_command and send it to
// the host.
//----------------------------------------------------------------------------
void send_packet(COMMAND_PACKET *packet)
  {
  ubyte
    i;
  unsigned char
     buffer[500];
  int
    index;
  index=0;
  buffer[index++]=packet->command;
  buffer[index++]=packet->data_length;
  for(i=0;i<packet->data_length;i++)
    buffer[index++]=packet->data[i];
  //Set the CRC
  packet->CRC.as_word=
    get_crc((ubyte *)packet,packet->data_length+2);
  buffer[index++]=packet->CRC.as_bytes[0];
  buffer[index++]=packet->CRC.as_bytes[1];

  DWORD
    bytes_written;
  bytes_written=0;

  if(hCom)
    if(TRUE!=WriteFile(hCom,              // handle of file to write to
                       buffer,            // address of data to write to file
                       index,             // number of bytes to write
                       &bytes_written,    // address of number of bytes written
                       NULL))
      ((CMy635_WinTestDlg*)OurParent)->m_comm_status.SetWindowText("WriteFile() failed!");
  }
//============================================================================
//                              check_for_packet()
//
// check_for_packet() will see if there is a valid packet in the input buffer.
// If there is, it will copy it into incoming_response and return 1. If there
// is not it will return 0. incoming_packet may get partially filled with
// garbage if there is not a valid packet available.
//----------------------------------------------------------------------------

int
  tossed=0;

ubyte check_for_packet(void)
  {
  //Clear the return packet.
  memset(&incoming_response,0,sizeof(COMMAND_PACKET));
  Sync_Read_Buffer();
  ubyte
    i;
  //First off, there must be at least 4 bytes available in the input stream
  //for there to be a valid command in it (command, length, no data, CRC).
  if(BytesAvail()<4)
    return(0);
  //The "peek" stuff allows us to look into the RS-232 buffer without
  //removing the data.
  Sync_Peek_Pointer();
  //Only commands 0 through MAX_COMMAND are valid.
  if(MAX_COMMAND<(0x3F&(incoming_response.command=PeekByte())))
    {
    //Throw out one byte of garbage. Next pass through should re-sync.
    GetByte();
    tossed++;
    return(0);
    }
  //There is a valid command byte. Get the data_length. The data length
  //must be within reason.
  if(MAX_DATA_LENGTH<(incoming_response.data_length=PeekByte()))
    {
    //Throw out one byte of garbage. Next pass through should re-sync.
    GetByte();
    tossed++;
    return(0);
    }
  //Now there must be at least incoming_response.data_length+sizeof(CRC) bytes
  //still available for us to continue.
  if((int)PeekBytesAvail()<(incoming_response.data_length+2))
    //It looked like a valid start of a packet, but it does not look
    //like the complete packet has been received yet.
    return(0);
  //There is enough data to make a packet. Transfer over the data.
  for(i=0;i<incoming_response.data_length;i++)
    incoming_response.data[i]=PeekByte();
  //Now move over the CRC.
  incoming_response.CRC.as_bytes[0]=PeekByte();
  incoming_response.CRC.as_bytes[1]=PeekByte();
  //Now check the CRC.
  if(incoming_response.CRC.as_word==
     get_crc((ubyte *)&incoming_response,incoming_response.data_length+2))
    {
    //This is a good packet. I'll be horn swaggled. Remove the packet
    //from the serial buffer.
    AcceptPeekedData();
    //Let our caller know that incoming_response has good stuff in it.
    return(1);
    }
  //The CRC did not match. Toss out one byte of garbage. Next pass through
  //should re-sync.
  GetByte();
  tossed++;
  return(0);
  }
//============================================================================